from typing import Any, Dict, List, Optional, Tuple, Type, Union

LiteralString = str


import json
import logging
from abc import ABC
from textwrap import indent

log: logging.Logger = logging.getLogger(__name__)

ChildType = Union["AbstractNode", str]
ITERABLES: Tuple[Type[Any], ...] = (list, tuple, set)


class QueryBuilder:
    method: str
    operation: str
    model: Optional[str]
    include: Optional[Dict[str, Any]]
    arguments: Dict[str, Any]
    root_selection: Optional[List[str]]

    __slots__ = (
        "method",
        "operation",
        "model",
        "include",
        "arguments",
        "root_selection",
    )

    def __init__(
        self,
        *,
        method: str,
        operation: str,
        arguments: Dict[str, Any],
        model: Optional[str] = None,
        root_selection: Optional[List[str]] = None,
    ) -> None:
        self.model = model
        self.method = method
        self.operation = operation
        self.root_selection = root_selection
        self.arguments = arguments
        self.include = arguments.pop("include", None)

    def build(self) -> str:
        data = {
            "variables": {},
            "operation_name": self.operation,
            "query": self.build_query(),
        }
        return dumps(data)

    def build_query(self) -> str:
        query = self._create_root_node().render()
        log.debug("Generated query: \n%s", query)
        return query

    def _create_root_node(self) -> "RootNode":
        root = RootNode(builder=self)
        root.add(ResultNode.create(self))
        root.add(
            Selection.create(
                self,
                model=self.model,
                include=self.include,
                root_selection=self.root_selection,
            )
        )
        return root


class AbstractNode(ABC):
    __slots__ = ()

    def should_render(self) -> bool:
        """If True, rendering of the node is skipped

        Useful for some nodes as they should only actually
        be rendered if they have any children.
        """
        return True


class Node(AbstractNode):
    """Base node handling rendering of child nodes"""

    joiner: str
    indent: str
    builder: QueryBuilder
    children: List[ChildType]

    __slots__ = (
        "joiner",
        "indent",
        "builder",
        "children",
    )

    def __init__(
        self,
        builder: QueryBuilder,
        *,
        joiner: str = "\n",
        indent: str = "  ",
        children: Optional[List[ChildType]] = None,
    ) -> None:
        self.builder = builder
        self.joiner = joiner
        self.indent = indent
        self.children = children if children is not None else []

    def render(self) -> Optional[str]:
        """Render the node and it's children and to string.

        Rendering a node involves 4 steps:

        1. Entering the node
        2. Rendering it's children
        3. Departing the node
        4. Joining the previous steps together into a single string
        """
        if not self.should_render():
            return None

        strings: List[str] = []
        entered = self.enter()
        if entered is not None:
            strings.append(entered)

        for child in self.children:
            content: Optional[str] = None

            if isinstance(child, str):
                content = child
            else:
                content = child.render()

            if content:
                strings.append(indent(content, self.indent))

        departed = self.depart()
        if departed is not None:
            strings.append(departed)

        return self.joiner.join(strings)

    def add(self, child: ChildType) -> None:
        """Add a child"""
        self.children.append(child)

    @classmethod
    def create(cls, builder: Optional[QueryBuilder] = None, **kwargs: Any) -> "Node":
        """Create the node and its children

        This is useful for subclasses that add extra attributes in __init__
        """
        kwargs.setdefault("builder", builder)
        node = cls(**kwargs)
        node.children.extend(node.create_children())
        return node


class RootNode(Node):
    __slots__ = ()

    def enter(self) -> str:
        return f"{self.builder.operation} {{"

    def depart(self) -> str:
        return "}"

    def render(self) -> str:
        content = super().render()
        return content


class ResultNode(Node):
    __slots__ = ()

    def __init__(self, indent: str = "", **kwargs: Any) -> None:
        super().__init__(indent=indent, **kwargs)

    def enter(self) -> str:
        # self.builder.model is None, model is empty string.
        model = self.builder.model or ""
        result = f"result: {self.builder.method}{model}"
        return result

    def depart(self) -> Optional[str]:
        return None

    def create_children(self) -> List[ChildType]:
        return [
            Arguments.create(
                self.builder,
                arguments=self.builder.arguments,
            )
        ]


class Arguments(Node):
    arguments: Dict[str, Any]

    __slots__ = ("arguments",)

    def __init__(self, arguments: Dict[str, Any], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.arguments = arguments

    def should_render(self) -> bool:
        return bool(self.children)

    def enter(self) -> str:
        return "("

    def depart(self) -> str:
        return ")"

    def create_children(
        self, arguments: Optional[Dict[str, Any]] = None
    ) -> List[ChildType]:
        children: List[ChildType] = []

        for arg, value in self.arguments.items():
            if isinstance(value, ITERABLES):
                children.append(f"{arg}: {dumps(dumps(value))}")
            else:
                children.append(f"{arg}: {dumps(value)}")

        return children


class Selection(Node):
    model: Optional[str]
    include: Optional[Dict[str, Any]]
    root_selection: Optional[List[str]]

    __slots__ = (
        "model",
        "include",
        "root_selection",
    )

    def __init__(
        self,
        model: Optional[str] = None,
        include: Optional[Dict[str, Any]] = None,
        root_selection: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.model = model
        self.include = include
        self.root_selection = root_selection

    def should_render(self) -> bool:
        return bool(self.children)

    def create_children(self) -> List[ChildType]:
        children: List[ChildType] = []
        return children


def dumps(obj: Any, **kwargs: Any) -> str:
    kwargs.setdefault("ensure_ascii", False)
    return json.dumps(obj, **kwargs)



import gzip
import logging
import os
import shutil
from pathlib import Path

import httpx
from pydantic import BaseModel

from .._config import check_extension, config
from .platform import binary_platform

log: logging.Logger = logging.getLogger()


class Engine(BaseModel):
    name: str = "query-engine"
    _session: httpx.Client = httpx.Client()
    _binary_path: str = f"{Path(__file__).parent.absolute()}/.binary"

    def download(self) -> None:
        # check .binary
        local_path = self._read_local_binary_path()
        if self._read_local_binary_path() and Path(local_path).exists():
            log.debug(f"{self.name} is cached, skipping download")
            return
        url = self.url
        dest = self.path
        # save .binary
        if dest.exists():
            log.debug(f"{self.name} is cached, skipping download")
            self._write_local_binary_path(to=str(dest.absolute()))
            return

        log.debug(f"Downloading from {url} to {dest}")
        self._download(url=url, to=str(dest.absolute()))
        log.debug(f"Downloaded {self.name} to {dest.absolute()}")

    def _download(self, url: str, to: str):
        Path(to).parent.mkdir(parents=True, exist_ok=True)

        tmp = to + ".tmp"
        tar = to + ".gz.tmp"

        with self._session.stream("GET", url, timeout=None) as resp:
            resp.raise_for_status()
            with open(tar, "wb") as fd:
                for chunk in resp.iter_bytes():
                    fd.write(chunk)

        # decompress to a tmp file before replacing the original
        with gzip.open(tar, "rb") as f_in:
            with open(tmp, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

        # chmod +x
        status = os.stat(tmp)
        os.chmod(tmp, status.st_mode | 0o111)

        # override the original
        shutil.copy(tmp, to)

        # remove temporary files
        os.remove(tar)
        os.remove(tmp)

        # save binary path
        self._write_local_binary_path(to=to)

    def _write_local_binary_path(self, to: str):
        with open(self._binary_path, mode="wb") as file:
            file.write(to.encode())

    def _read_local_binary_path(self) -> str:
        try:
            with open(self._binary_path, mode="rb") as file:
                data = file.read().decode()
            return data
        except:
            ...

    @property
    def binary_path(self) -> str:
        path = self._read_local_binary_path()
        if path:
            return path
        self.download()
        return str(self.path)

    @property
    def url(self) -> str:
        return check_extension(
            config.engine_url.format(config.engine_version, binary_platform(), self.name)
        )

    @property
    def path(self) -> Path:
        binary_name = binary_platform()
        return Path(
            check_extension(
                str(config.global_temp_dir.joinpath(f"prisma-{self.name}-{binary_name}"))
            )
        )



import asyncio
import json
import logging
import subprocess
from os import environ
from pathlib import Path
from typing import Any, Dict, List

import httpx

from .._binary import check_binary
from .._config import config
from . import common
from .errors import (  # UnprocessableEntityError,
    AlreadyConnectedError,
    BaseStartEngineError,
    EngineConnectionError,
    EngineRequestError,
    NotConnectedError,
    StartEngineError,
    handler_error,
)

log: logging.Logger = logging.getLogger()


class AsyncEngine:
    __slots__ = (
        "db_uri",
        "db_provider",
        "db_timeout",
        "connect_timeout",
        "url",
        "process",
        "session",
        "connected",
        "_binary_path",
    )

    def __init__(
        self,
        db_uri: str,
        db_provider: str,
        db_timeout: int = 10,
        connect_timeout: int = 10,
        url: str = None,
        process: subprocess.Popen = None,
        session: httpx.AsyncClient = None,
    ) -> None:
        self.db_uri = db_uri
        self.db_provider = db_provider
        self.db_timeout = db_timeout
        self.connect_timeout = connect_timeout

        self.url: str = url
        self.process: subprocess.Popen = process
        self.session: httpx.AsyncClient = session
        self.connected = False
        self._binary_path: Path = None

    def __del__(self):
        if self.process:
            self.url = None
            self.process.kill()
            self.process = None
            self.connected = False
            self.session = None

    async def _close(self):
        if self.session and not self.session.is_closed:
            await self.session.aclose()
        self.session = None

    async def connect(self) -> None:
        if self.process:
            raise AlreadyConnectedError("Already connected to the engine")

        self.session: httpx.AsyncClient = httpx.AsyncClient(timeout=self.db_timeout)
        self._binary_path = check_binary()

        try:
            await self.spawn(file=Path(self._binary_path))
        except Exception:
            await self._close()
            raise

    async def disconnect(self) -> None:
        await self._close()
        if not self.process:
            raise NotConnectedError("Not connected")

        self.url = None
        self.process.kill()
        self.process = None
        self.connected = False

    async def _try_comunicate(self):
        try:
            stdout, stderr = self.process.communicate(timeout=0.03)

            stderr = stderr.decode()
            stdout = stdout.decode()

            std = stdout or stderr

            try:
                data = json.loads(std)
                await self.disconnect()
                raise StartEngineError(error=data)
            except (json.JSONDecodeError, TypeError):
                await self.disconnect()
                raise BaseStartEngineError(std)
        except subprocess.TimeoutExpired:
            pass

    async def spawn(self, file: Path) -> None:
        port = common.get_open_port()
        log.debug(f"Running engine on port: {port}")

        self.url = f"http://localhost:{port}"

        dml = await common.get_dml()

        dml = dml.replace("postgres", self.db_provider)
        dml = dml.replace("DATASOURCE_URL", self.db_uri)

        env = environ.copy()

        env.update(
            PRISMA_DML=dml,
            RUST_LOG="error",
            RUST_LOG_FORMAT="json",
            PRISMA_CLIENT_ENGINE_TYPE="binary",
        )

        args: List[str] = [str(file.absolute()), "-p", str(port), "--enable-raw-queries"]

        log.debug("Starting engine...")
        self.process = subprocess.Popen(
            args,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        await self._try_comunicate()

        await self._check_connect()

    async def request(
        self, method: config.method, path: str, *, content: Any = None
    ) -> Dict[str, Any]:
        if not self.url:
            raise NotConnectedError("Not connected to the engine")

        kwargs = {
            "headers": {
                "Content-Type": "application/json",
                "Accept": "application/json",
            }
        }

        if content:
            kwargs["content"] = content

        url = self.url + path
        log.debug(f"Sending {method} request to {url} with content: {content}")

        resp = await self.session.request(method=method, url=url, **kwargs)

        if resp.is_success:
            response = resp.json()

            log.debug(f"{method} {url} returned {len(response)} results")

            errors_data = response.get("errors")
            if errors_data:
                return handler_error(errors=errors_data)

            return response

        # unused
        # elif resp.status == 422:
        #    raise UnprocessableEntityError(resp)

        raise EngineRequestError(f"{resp.status_code}: {resp.text}")

    async def _check_connect(self) -> None:
        last_err = None
        for _ in range(int(self.connect_timeout / 0.01)):
            try:
                await self.request("GET", "/status")
                self.connected = True
                return
            except Exception as err:
                await asyncio.sleep(0.01)
                log.debug(
                    (
                        "Could not connect to engine due to "
                        f"{err.__class__.__name__}; retrying...{err}"
                    )
                )
                last_err = err

        await self._try_comunicate()

        await self.disconnect()

        raise EngineConnectionError("Could not connect to the engine") from last_err



from typing import Dict, Type

from ._base_error import SQLXEngineError


class UniqueViolationError(SQLXEngineError):
    ...


class ForeignKeyViolationError(SQLXEngineError):
    ...


class FieldNotFoundError(SQLXEngineError):
    ...


class RawQueryError(SQLXEngineError):
    ...


class MissingRequiredValueError(SQLXEngineError):
    ...


class InputError(SQLXEngineError):
    ...


class TableNotFoundError(SQLXEngineError):
    ...


class RecordNotFoundError(SQLXEngineError):
    ...


ERRORS_MAPPING: Dict[str, Type[Exception]] = {
    "P2002": UniqueViolationError,
    "P2003": ForeignKeyViolationError,
    "P2009": FieldNotFoundError,
    "P2010": RawQueryError,
    "P2012": MissingRequiredValueError,
    "P2019": InputError,
    "P2021": TableNotFoundError,
    "P2025": RecordNotFoundError,
}



from datetime import date, datetime, time
from decimal import Decimal
from typing import Any, Dict, List, Tuple
from uuid import UUID

from pydantic import Json, create_model, parse_obj_as

from .common import BaseRow

TYPES = {
    "int": int,
    "bigint": int,
    "float": float,
    "double": float,
    "string": str,
    "bool": bool,
    "char": str,
    "decimal": Decimal,
    "json": Json,
    "uuid": UUID,
    "datetime": datetime,
    "date": date,
    "time": time,
    "array": list,
    "xml": str,
    # not implement
    "bytes": Any,  # bytes
    "enum": Any,  # Enum
    "null": Any,
}


class Deserialize:
    __slots__ = ("rows", "as_base_row", "_model", "_base_model_type")

    def __init__(self, rows: List[Dict[str, Any]], as_base_row: bool = True) -> None:
        self.rows = rows
        self.as_base_row = as_base_row
        self._model: BaseRow = None
        self._base_model_type: dict = {}

    def deserialize(self):
        if self.as_base_row:
            self._create_base_types(self.rows[0])
        mapper = map(self._get_row, self.rows)
        if not self.as_base_row:
            return mapper
        return map(self._to_model, mapper)

    def _to_model(self, row: dict) -> BaseRow:
        return self._model(**row)

    def _create_base_types(self, first_row: dict):
        for key in first_row.keys():
            _type = TYPES.get(first_row[key]["prisma__type"], Any)
            self._base_model_type.update({key: (_type, None)})
        self._create_model()

    def _create_model(self) -> None:
        _model = create_model("BaseRow", **self._base_model_type)
        self._model = _model

    def _get_row(self, row: dict):
        mapper = {}
        for key in row.keys():
            mapper.update({key: row[key]["prisma__value"]})
        return mapper



import logging
from functools import lru_cache
from time import time

from httpx import HTTPStatusError

from .engine import Engine

log: logging.Logger = logging.getLogger()


@lru_cache(maxsize=None)
def check_binary(*args, **kwargs) -> str:
    log.debug("Checking engine binary file.")
    ini = time()
    try:
        engine = Engine()
        path = engine.binary_path
        log.debug(f"Check binary ok. time: {time() - ini}")
        return path
    except HTTPStatusError:
        log.error("Error: Download of the query binary.")
        raise



import socket
from pathlib import Path

import aiofiles
from pydantic import BaseModel


def get_open_port() -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return int(port)


async def get_dml() -> str:
    path = f"{Path(__file__).parent.absolute()}/schema.prisma"
    async with aiofiles.open(path, mode="r") as file:
        content = await file.read()
    return content


class BaseRow(BaseModel):
    ...



import json
from typing import Any, Dict, List

from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import JsonLexer

from .._config import config
from ._base_error import GenericSQLXEngineError, SQLXEngineError
from ._errors_mapping import COMMON_ERRORS_MAPPING, ENGINE_ERRORS_MAPPING
from ._except_mapping import ERRORS_MAPPING


class AlreadyConnectedError(SQLXEngineError):
    ...


class EngineConnectionError(SQLXEngineError):
    ...


class NotConnectedError(SQLXEngineError):
    ...


class EngineRequestError(SQLXEngineError):
    ...


# class UnprocessableEntityError(SQLXEngineError):
#    def __init__(self, resp: httpx.Response, *args) -> None:
#        error_msg = f"status_code: {resp.status_code} -> body: {resp.json()}"
#        super().__init__(error_msg, *args)


class EngineError(SQLXEngineError):
    ...


class BaseStartEngineError(SQLXEngineError):
    ...


class StartEngineError(BaseStartEngineError):
    def __init__(self, error: dict, *args: object) -> None:
        error_code = error.get("error_code")
        message = error.get("message", "")
        is_panic = error.get("is_panic", False)
        meta = error.get("meta", None)
        error_msg = COMMON_ERRORS_MAPPING.get(error_code)
        complete_msg = _mount_msg(
            error_code=error_code,
            message=message,
            error_msg=error_msg,
            meta=meta,
            is_panic=is_panic,
        )
        super().__init__(complete_msg)


class SQLXEngineTimeoutError(SQLXEngineError):
    ...


def _colorizer_json(dumps: str):
    return highlight(dumps, JsonLexer(), TerminalFormatter())


def _mount_msg(
    error_code: str, message: str, error_msg: str, meta: dict, is_panic: bool
):
    error_msg = json.dumps(
        {
            "is_panic": is_panic,
            "error_code": error_code,
            "error_message": message,
            "meta": meta,
            "helper": (
                "https://www.prisma.io/docs/reference/"
                "api-reference/error-reference#error-codes"
            ),
            "description": error_msg,
        },
        indent=2 if config.improved_error_log else None,
        sort_keys=False,
    )
    if not config.improved_error_log:
        return "\n" + error_msg
    return "\n" + _colorizer_json(error_msg)


def handler_error(errors: List[Dict[str, Any]]):

    for error in errors:
        user_facing_error = error.get("user_facing_error", {})

        error_code = user_facing_error.get("error_code")
        message = user_facing_error.get("message", "")
        is_panic = user_facing_error.get("is_panic", False)
        meta = user_facing_error.get("meta", False)

        error_msg = ENGINE_ERRORS_MAPPING.get(error_code)

        complete_msg = _mount_msg(
            error_code=error_code,
            message=message,
            error_msg=error_msg,
            meta=meta,
            is_panic=is_panic,
        )
        error_mapping = ERRORS_MAPPING.get(error_code)
        if error_mapping:
            raise error_mapping(complete_msg)
        raise SQLXEngineError(complete_msg)

    raise GenericSQLXEngineError(
        f"Could not process erroneous response: {json.dumps(errors, indent=2)}"
    )



from types import TracebackType
from typing import Dict, List, Literal, Optional, Type, Union

import httpx
from typing_extensions import LiteralString

from ._config import config
from ._core.builder import QueryBuilder
from ._core.common import BaseRow
from ._core.engine import AsyncEngine as _AsyncEngine
from ._core.errors import (
    AlreadyConnectedError,
    NotConnectedError,
    SQLXEngineTimeoutError,
)
from ._core.parser import Deserialize


class SQLXEngine:
    """
    SQLXEngine is an engine to run pure sql with queries, inserts,
    deletes, updates, create/alter/drop tables etc.

    All SQL that is executed using the SQLXEngine is atomic; that is,
    only one statement is performed at a time. Only the first one will
    be completed if you send an Insert and a select.
    This is one of the ways to deal with SQL ingestion.

    Always `COMMIT` and `ROLLBACK` is automatic!!! This is not changeable...


    `providers(str)`: postgresql, mysql, sqlserver or sqlite
    `uri(str)`:  Connection values.
    `timeout(int)`: Default 10 sec, used to close the connection after this time. Or None to removed timeout.
    `improved_error_log(bool)`: Default True. Returns an error in json format with colors.

    Usage:
    ``` python
    >>> ##### PostgreSQL
    >>> uri = "postgresql://user:pass@host:port/db?schema=sample"
    >>> conn = SQLXEngine(provider="postgresql", uri=uri)

    #------------------------------------------

    >>> ##### MySQL
    >>> uri = "mysql://user:pass@host:port/db?schema=sample"
    >>> conn = SQLXEngine(provider="mysql", uri=uri)

    #------------------------------------------

    >>> ##### Microsoft SQL Server
    >>> uri = "sqlserver://host:port;initial catalog=sample;user=sa;password=pass;"
    >>> conn = SQLXEngine(provider="sqlserver", uri=uri)

    #------------------------------------------

    >>> ##### SQLite
    >>> uri = "file:./dev.db"
    >>> conn = SQLXEngine(provider="sqlite", uri=uri)

    # URI Reference: https://www.prisma.io/docs/reference/database-reference/connection-urls
    ```
    """

    __slots__ = ("provider", "uri", "timeout", "connected", "_connection")

    def __init__(
        self,
        provider: Literal["postgresql", "mysql", "sqlserver", "sqlite"],
        uri: str,
        timeout: int = 10,
        improved_error_log: bool = True,
    ) -> None:
        _providers = ["postgresql", "mysql", "sqlserver", "sqlite"]
        if provider not in _providers:
            raise ValueError(
                f"Invalid provider {provider} \n Providers available: {_providers}"
            )
        if not uri or not any([uri.startswith(prov) for prov in [*_providers, "file"]]):
            raise ValueError(f"Invalid uri: {uri}, check the usage uri.")

        self.uri = uri
        self.provider = provider
        self.connected: bool = False
        self.timeout: int = timeout
        self._connection: _AsyncEngine = None

        config.improved_error_log = improved_error_log

    async def __aenter__(self) -> "SQLXEngine":
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        if self._connection:
            await self.close()

    async def connect(self, timeout: int = 10) -> None:
        """
        Every connection instance is lazy, only after the connect is the
        database checked and a connection is created with it.

        Args:
            timeout in sec (int): time max to try connect with db.
            You can change the value if your database is slow to receive a new connection
            Defaults to 10 sec.
        """
        if not isinstance(timeout, (int, float)):
            raise ValueError("Invalid timeout, timeout must be a number.")
        if self._connection:
            raise AlreadyConnectedError("Already connected to the engine")
        self._connection = _AsyncEngine(
            db_uri=self.uri,
            db_provider=self.provider,
            db_timeout=self.timeout,
            connect_timeout=timeout,
        )
        await self._connection.connect()
        self.connected = self._connection.connected

    async def close(self) -> None:
        """Is good always close the connection, but!
        Even if you don't close the connection, don't worry,
        when the process ends automatically the connections will
        be closed so the bank doesn't have an idle connection.
        """
        if not self._connection:
            raise NotConnectedError("Not connected")
        await self._connection.disconnect()
        self.connected = False
        self._connection = None

    async def execute(self, stmt: LiteralString) -> int:
        """Execute statement to change, add or delete etc.
        ---
        Always `COMMIT` and `ROLLBACK` is automatic!!! This is not changeable...
        if you transaction failed your receive a except with error.
        ---
        - `Exception Example`:

        ``` python
            sqlx_engine.core.errors.RawQueryError:
            {
                "is_panic": false,
                "error_code": "P2010",
                "error_message": "Raw query failed. Code: `23505`. Message: `Key (id)=(a7e382c9-8d6d-4233-b1be-be9ef6024bd5) already exists.`",
                "meta": {
                    "code": "23505",
                    "message": "Key (id)=(a7e382c9-8d6d-4233-b1be-be9ef6024bd5) already exists."
                },
                "helper": "https://www.prisma.io/docs/reference/api-reference/error-reference#error-codes",
                "description": "Raw query failed. Code: {code}. Message: {message}"
            }

        ```
        ---
        - `Args`:
            * `stmt (LiteralString)`: Insert, Update, Delete, Drop etc. `Query NO`!

        ---
        - `Usage`:

            >>> await db.execute(query="INSERT INTO table(name) values ('rian')")
            >>> await db.execute(query="INSERT INTO table(name) values (?)")

        ---
        - `Returns`:
            * `int`: Number of rows affected

        ---
        - `Raises`:
            * `RawQueryError`: Default
            * `SQLXEngineError`
            * `SQLXEngineTimeoutError`
            * `GenericSQLXEngineError`
        """
        if not self._connection:
            raise NotConnectedError("Not connected")

        builder = QueryBuilder(
            method="executeRaw",
            operation="mutation",
            arguments={
                "query": stmt,
                "parameters": [],
            },
        )
        content = builder.build()
        try:
            data = await self._connection.request(
                method="POST", path="/", content=content
            )
        except (httpx.TimeoutException):
            raise SQLXEngineTimeoutError(
                f"Error on .execute - Timeout after: {self.timeout} secs"
            )
        return int(data["data"]["result"])

    async def query(
        self,
        query: LiteralString,
        as_base_row: bool = True,
    ) -> Optional[Union[List[BaseRow], List[Dict]]]:
        """Execute query on db
        ---
        - `Args`:
            * `query (LiteralString)`: for example, `SELECT * FROM user` to return actual records.
            * `as_base_row (bool)`: By default is True, BaseRow is an object typing.

        ---
        - `Usage`:

            >>> await db.query(query="SELECT 1")

        ---
        - `Returns`:
            * `List[BaseRow]`: List of Pydantic Model
            * `List[Dict]`: List of dict
            * `NoneType`: None

        ---
        - `Raises`:
            * `RawQueryError` Default
            * `SQLXEngineError`
            * `SQLXEngineTimeoutError`
            * `GenericSQLXEngineError`

        """
        if not self._connection:
            raise NotConnectedError("Not connected")
        builder = QueryBuilder(
            method="queryRaw",
            operation="mutation",
            arguments={
                "query": query,
                "parameters": [],
            },
        )
        content = builder.build()
        try:
            data = await self._connection.request(
                method="POST", path="/", content=content
            )
        except (httpx.TimeoutException):
            raise SQLXEngineTimeoutError(
                f"Error on .query - Timeout after: {self.timeout} secs"
            )
        resp = data["data"]["result"]
        if resp:
            rows = Deserialize(rows=resp, as_base_row=as_base_row)
            return list(rows.deserialize())
        return None



from ._core._base_error import GenericSQLXEngineError as GenericSQLXEngineError
from ._core._base_error import SQLXEngineError as SQLXEngineError
from ._core._except_mapping import FieldNotFoundError as FieldNotFoundError
from ._core._except_mapping import ForeignKeyViolationError as ForeignKeyViolationError
from ._core._except_mapping import InputError as InputError
from ._core._except_mapping import (
    MissingRequiredValueError as MissingRequiredValueError,
)
from ._core._except_mapping import RawQueryError as RawQueryError
from ._core._except_mapping import RecordNotFoundError as RecordNotFoundError
from ._core._except_mapping import TableNotFoundError as TableNotFoundError
from ._core._except_mapping import UniqueViolationError as UniqueViolationError
from ._core.errors import AlreadyConnectedError as AlreadyConnectedError
from ._core.errors import NotConnectedError as NotConnectedError
from ._core.errors import SQLXEngineTimeoutError as SQLXEngineTimeoutError



import re
import subprocess
import sys
from functools import lru_cache
from typing import Tuple

from .._config import config


@lru_cache(maxsize=None)
def binary_platform(*args, **kwargs) -> str:
    if config.platform_name != "linux":
        return config.platform_name

    distro = linux_distro()
    ssl = get_openssl()

    return "linux-musl" if distro == "alpine" else f"{distro}-openssl-{ssl}"


def linux_distro() -> str:
    dist_id, dist_like = _get_linux_distro_details()
    types = {
        "alpine": "alpine",
        "rhel": "rhel",
    }
    _name = any([_d in dist_like for _d in ["centos", "fedora", "rhel"]])
    distro = types.get(dist_id) or types.get("rhel" if _name else "")

    return distro or "debian"  # default debian


def _get_linux_distro_details() -> Tuple[str, str]:
    process = subprocess.run(
        ["cat", "/etc/os-release"], stdout=subprocess.PIPE, check=True
    )
    output = str(process.stdout, sys.getdefaultencoding())

    match = re.search(r'ID="?([^"\n]*)"?', output)
    distro_id = match.group(1) if match else ""  # type: str

    match = re.search(r'ID_LIKE="?([^"\n]*)"?', output)
    distro_id_like = match.group(1) if match else ""  # type: str
    return distro_id, distro_id_like


def get_openssl() -> str:
    process = subprocess.run(
        ["openssl", "version", "-v"], stdout=subprocess.PIPE, check=True
    )
    return parse_openssl_version(str(process.stdout, sys.getdefaultencoding()))


def parse_openssl_version(string: str) -> str:
    match = re.match(r"^OpenSSL\s(\d+\.\d+)\.\d+", string)
    return "1.1.x" if match is None else match.group(1) + ".x"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sqlx_engine',
 'sqlx_engine._binary',
 'sqlx_engine._config',
 'sqlx_engine._core']

package_data = \
{'': ['*']}

install_requires = \
['Pygments>=2.12.0,<3.0.0',
 'aiofiles>=0.8.0,<0.9.0',
 'httpx>=0.23.0,<0.24.0',
 'pydantic>=1.9.1,<2.0.0']

setup_kwargs = {
    'name': 'pysqlx-engine',
    'version': '0.1.14',
    'description': 'Python Async SQL Engines',
    'long_description': '# PySQLXEngine\n\n<p align="center">\n  <a href="/"><img src="https://carlos-rian.github.io/pysqlx-engine/img/logo-text3.png" alt="PySQLXEngine Logo"></a>\n</p>\n<p align="center">\n    <em>PySQLXEngine, a minimalist asynchronous SQL engine</em>\n</p>\n\n<p align="center">\n<a href="https://github.com/carlos-rian/pysqlx-engine/actions?query=workflow%3ATest+event%3Apush+branch%3Amain" target="_blank">\n    <img src="https://github.com/carlos-rian/pysqlx-engine/workflows/Test/badge.svg?event=push&branch=main" alt="Test">\n</a>\n<a href="https://app.codecov.io/gh/carlos-rian/pysqlx-engine" target="_blank">\n    <img src="https://img.shields.io/codecov/c/github/carlos-rian/pysqlx-engine?color=%2334D058" alt="Coverage">\n</a>\n<a href="https://pypi.org/project/pysqlx-engine" target="_blank">\n    <img src="https://img.shields.io/pypi/v/pysqlx-engine?color=%2334D058&label=pypi%20package" alt="Package version">\n</a>\n<a href="https://pypi.org/project/pysqlx-engine" target="_blank">\n    <img src="https://img.shields.io/pypi/pyversions/pysqlx-engine.svg?color=%2334D058" alt="Supported Python versions">\n</a>\n</p>\n\n---\n\n**Documentation**: <a href="https://carlos-rian.github.io/pysqlx-engine/" target="_blank">https://carlos-rian.github.io/pysqlx-engine/</a>\n\n**Source Code**: <a href="https://github.com/carlos-rian/pysqlx-engine" target="_blank">https://github.com/carlos-rian/pysqlx-engine</a>\n\n---\n\nPySQLXEngine supports the option of sending **raw sql** to your database.\n\nThe PySQLXEngine is a minimalist **Async** SQL engine. Currently this lib only supports *asynchronous programming*, you need to use `await` in all methods.\n\nDatabase Support:\n\n* `SQLite`\n* `PostgreSQL`\n* `MySQL`\n* `Microsoft SQL Server`\n\nOS Support:\n\n* `Linux`\n* `Windows` *Experimental! Unit tests were not run on Windows.*\n\n## Installation\n\n\nPIP\n\n```console\n$ pip install pysqlx-engine\n```\n\nPoetry\n\n```console\n$ poetry add pysqlx-engine\n```\n\n\n## Example\n\n* Create `main.py` file.\n\n```python\nimport asyncio\n\nfrom sqlx_engine import SQLXEngine\n\nuri = "file:./db.db"\ndb = SQLXEngine(provider="sqlite", uri=uri)\n\nasync def main():\n    await db.connect()\n    rows = await db.query(query="select 1 as number")\n    print(rows)\n\nasyncio.run(main())\n```\n\n* Run it\n\n<div class="termy">\n\n```console\n$ python3 main.py\n\n[BaseRow(number=1)]\n```\n</div>',
    'author': 'Carlos Rian',
    'author_email': 'crian.rian@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://carlos-rian.github.io/pysqlx-engine',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



ENGINE_ERRORS_MAPPING = {
    "P2000": "The provided value for the column is too long for the column's type. Column: {column_name}",
    "P2001": "The record searched for in the where condition ({model_name}.{argument_name} = {argument_value}) does not exist",
    "P2002": "Unique constraint failed on the {constraint}",
    "P2003": "Foreign key constraint failed on the field: {field_name}",
    "P2004": "A constraint failed on the database: {database_error}",
    "P2005": "The value {field_value} stored in the database for the field {field_name} is invalid for the field's type",
    "P2006": "The provided value {field_value} for {model_name} field {field_name} is not valid",
    "P2007": "Data validation error {database_error}",
    "P2008": "Failed to parse the query {query_parsing_error} at {query_position}",
    "P2009": "Failed to validate the query: {query_validation_error} at {query_position}",
    "P2010": "Raw query failed. Code: {code}. Message: {message}",
    "P2011": "Null constraint violation on the {constraint}",
    "P2012": "Missing a required value at {path}",
    "P2013": "Missing the required argument {argument_name} for field {field_name} on {object_name}.",
    "P2014": "The change you are trying to make would violate the required relation '{relation_name}' between the {model_a_name} and {model_b_name} models.",
    "P2015": "A related record could not be found. {details}",
    "P2016": "Query interpretation error. {details}",
    "P2017": "The records for relation {relation_name} between the {parent_name} and {child_name} models are not connected.",
    "P2018": "The required connected records were not found. {details}",
    "P2019": "Input error. {details}",
    "P2020": "Value out of range for the type. {details}",
    "P2021": "The table {table} does not exist in the current database.",
    "P2022": "The column {column} does not exist in the current database.",
    "P2023": "Inconsistent column data: {message}",
    "P2024": "Timed out fetching a new connection from the connection pool. (More info: http://pris.ly/d/connection-pool, Current connection limit: {connection_limit})",
    "P2025": "An operation failed because it depends on one or more records that were required but not found. {cause}",
    "P2026": "The current database provider doesn't support a feature that the query used: {feature}",
    "P2027": "Multiple errors occurred on the database during query execution: {errors}",
    "P2030": "Cannot find a fulltext index to use for the search, try adding a @@fulltext([Fields...]) to your schema",
    "P2033": "A number used in the query does not fit into a 64 bit signed integer. Consider using BigInt as field type if you're trying to store large integers",
}

# reference -> https://www.prisma.io/docs/reference/api-reference/error-reference#error-codes
COMMON_ERRORS_MAPPING = {
    "P1000": "Authentication failed against database server at {database_host}, the provided database credentials for {database_user} are not valid. Please make sure to provide valid database credentials for the database server at {database_host}.",
    "P1001": "Can't reach database server at {database_host}:{database_port} Please make sure your database server is running at {database_host}:{database_port}.",
    "P1002": "The database server at {database_host}:{database_port} was reached but timed out. Please try again. Please make sure your database server is running at {database_host}:{database_port}.",
    "P1003": (
        "Database {database_file_name} does not exist at {database_file_path}\n"
        "Database {database_name}.{database_schema_name} does not exist on the database server at {database_host}:{database_port}.\n"
        "Database {database_name} does not exist on the database server at {database_host}:{database_port}."
    ),
    "P1008": "Operations timed out after {time}",
    "P1009": "Database {database_name} already exists on the database server at {database_host}:{database_port}",
    "P1010": "User {database_user} was denied access on the database {database_name}",
    "P1011": "Error opening a TLS connection: {message}",
    "P1012": (
        "Possible errors:\n"
        "Argument {} is missing.\n"
        "Function {} takes {} arguments, but received {}.\n"
        "Argument {} is missing in attribute @{}.\n"
        "Argument {} is missing in data source block {}.\n"
        "Argument {} is missing in generator block {}.\n"
        "Error parsing attribute @{}: {}\n"
        "Attribute @{} is defined twice.\n"
        "The model with database name {} could not be defined because another model with this name exists: {}\n"
        "{} is a reserved scalar type name and can not be used.\n"
        "The {} {} cannot be defined because a {} with that name already exists.\n"
        "Key {} is already defined in {}.\n"
        "Argument {} is already specified as unnamed argument.\n"
        "Argument {} is already specified.\n"
        "No such argument.\n"
        "Field {} is already defined on model {}.\n"
        "Field {} in model {} can't be a list. The current connector does not support lists of primitive types.\n"
        "The index name {} is declared multiple times. With the current connector index names have to be globally unique.\n"
        "Value {} is already defined on enum {}.\n"
        "Attribute not known: @{}.\n"
        "Function not known: {}.\n"
        "Datasource provider not known: {}.\n"
        "shadowDatabaseUrl is the same as url for datasource {}. Please specify a different database as shadow database.\n"
        "The preview feature {} is not known. Expected one of: {}\n"
        "{} is not a valid value for {}.\n"
        "Type {} is neither a built-in type, nor refers to another model, custom type, or enum.\n"
        "Type {} is not a built-in type.\n"
        "Unexpected token. Expected one of: {}\n"
        "Environment variable not found: {}.\n"
        "Expected a {} value, but received {} value {}.\n"
        "Expected a {} value, but failed while parsing {}: {}.\n"
        "Error validating model {}: {}\n"
        "Error validating field {} in model {}: {}\n"
        "Error validating datasource {datasource}: {message}\n"
        "Error validating enum {}: {}\n"
        "Error validating: {}"
    ),
    "P1013": "The provided database string is invalid. {details}",
    "P1014": "The underlying {kind} for model {model} does not exist.",
    "P1015": (
        "Your Prisma schema is using features that are not supported for the version of the database."
        "Database version: {database_version}"
    ),
    "P1016": "Your raw query had an incorrect number of parameters. Expected: {expected}, actual: {actual}.",
    "P1017": "Server has closed the connection.",
}



import platform
import tempfile
from pathlib import Path
from typing import Literal

from pydantic import BaseConfig


class Config(BaseConfig):
    improved_error_log: bool = True
    platform_name: str = platform.system().lower()
    method: Literal = Literal["GET", "POST"]
    engine_version: str = "9f27eed0f8b8773287ff09e657bfa267983e5948"
    # engine_version: str = "81a0ee489e5f8c8ce98440028ea9050092430503"
    engine_url: str = "https://binaries.prisma.sh/all_commits/{0}/{1}/{2}.gz"
    global_temp_dir: Path = (
        Path(tempfile.gettempdir()) / "prisma" / "binaries" / "engines" / engine_version
    )


config = Config()


def check_extension(filename: str) -> str:
    if config.platform_name == "windows" and ".exe" not in filename:
        if ".gz" in filename:
            return filename.replace(".gz", ".exe.gz")
        return filename + ".exe"
    return filename
